﻿CompilerIf Defined(INCLUDE_COMDOC, #PB_Constant)=0
#INCLUDE_COMDOC=1
;/////////////////////////////////////////////////////////////////////////////////
;***comDoc***  - Version 1.3.
;============
;   Compound documents (structured storage).
;
;   Created with Purebasic 4.5 for Windows.
;
;   Platforms:  Windows 2000 onwards (x86 and x64).
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;NOTES :
;
;   i)    This library uses OLE's compound file implementation of the iStorage and iStream interfaces which does come
;         with some limitations, the main one being that we cannot have multiple simultaneous access to a file object.
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-CONSTANTS / STRUCTURES / PROTOTYPES.

CompilerIf Defined(statstg, #PB_Structure) = 0
    Structure STATSTG
      pwcsName.i        ;Unicode string.
      type.i
      cbSize.q
      mtime.FILETIME
      ctime.FILETIME
      atime.FILETIME
      grfMode.i
      grfLocksSupported.i
      clsid.CLSID
      grfStateBits.i
      reserved.i
    EndStructure 
  CompilerEndIf


  Structure comDocDocumentObject_ClassTemplate
    *vTable.i
    fileName$
    storage.IStorage
    enumerate.IEnumSTATSTG
    malloc.IMalloc
    statstg.STATSTG
    lockbytes.ILockBytes
  EndStructure

  Structure comDocFileObject_ClassTemplate
    *vTable.i
    name$
    stream.IStream
    errorFlag.i
  EndStructure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-DECLARES.
  Declare comDoc_documentClass_FinishEnumeration(*this.comDocDocumentObject_ClassTemplate)
  Declare.i comDoc_fileClass_Loc(*this.comDocFileObject_ClassTemplate)
  Declare.i comDoc_fileClass_Lof(*this.comDocFileObject_ClassTemplate)
  Declare comDoc_fileClass_WriteWord(*this.comDocFileObject_ClassTemplate, value.w)
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-INTERFACES.

  Interface comDocDocumentObject
    Close()
    CreateFile.i(name$)             ;Replaces any existing file of the same name.  
                                    ;Returns a comDocFileObject pointer if successful.
    DeleteFile.i(name$)             ;Returns non-zero if successful.
    OpenFile.i(name$)               ;Returns a comDocFileObject pointer if successful.
    SaveDocumentAs.i(fileName$, blnUseNewDocument=#False)
                                    ;Save to disc file only. (Cannot save to memory.)
                                    ;Over-writes any existing disc file with a snapshot of the underlying document.
                                    ;fileName$ = "" for a temporary disc-based document.
                                    ;Set blnUseNewDocument to non-zero to have the underlying document object reference the
                                    ;newly saved document. In these cases any open comDocFileObject objects will be invalidated.
                                    ;Returns non-zero if successful.
    ;Enumeration methods.
      BeginEnumeration.i()          ;Returns #True if successful.
      EnumerationEntryName.s()
      EnumerationEntrySize.i()
      EnumerationEntryType.i()      ;Returns #STGTY_STORAGE (folder) or #STGTY_STREAM (file) or some other value.
      FinishEnumeration()
      NextEnumerationEntry.i()      ;Returns #True if we are not already at the end of the enumeration.
  EndInterface

  Interface comDocFileObject
    Close()
    ;General file related methods.
      Eof.i()                                   ;Returns #True or #False.
      FileSeek(newPosition)
      Loc.i()
      Lof.i()
    ;Read methods.
      ReadAsciiCharacter.a()
      ReadByte.b()
      ReadCharacter.c()
      ReadData.i(*memoryBuffer, lengthToRead)   ;Returns the number of bytes actually read.
      ReadDouble.d()
      ReadFloat.f()
      ReadInteger.i()
      ReadLong.l()
      ReadQuad.q()
      ReadString.s(format=0)
      ReadUnicodeCharacter.u()
      ReadWord.w()
    ;Write methods.
      WriteAsciiCharacter(value.a)
      WriteByte(value.b)
      WriteCharacter(value.c)
      WriteData(*memoryBuffer, length)
      WriteDouble(value.d)
      WriteFloat(value.f)
      WriteInteger(value.i)
      WriteLong(value.l)
      WriteQuad(value.q)
      WriteString(text$, format=0)              ;A null terminator is added. If format = 0, then the string is written in Ascii/Unicode
                                                ;depending on the Unicode compiler switch etc.
      WriteUnicodeCharacter(value.u)
      WriteWord(value.w)
    ;Miscellaneous methods.
      ClearErrorFlag()
      GetErrorFlag.i()
  EndInterface
;/////////////////////////////////////////////////////////////////////////////////



;-INTERNAL FUNCTIONS.

;/////////////////////////////////////////////////////////////////////////////////
;The following frees a unicode string.
Procedure comDoc_INTERNAL_FreeUnicodeString(buffer)
  CompilerIf #PB_Compiler_Unicode = 0
    FreeMemory(buffer)
  CompilerEndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following returns the address of a Unicode character array containing a copy of the given string.
Procedure.i comDoc_INTERNAL_StringToUnicode(*ptrChar.CHARACTER)
  Protected result, *str.STRING
  CompilerIf #PB_Compiler_Unicode
    result = *ptrChar
  CompilerElse
    *str = @*ptrChar
    result = AllocateMemory(Len(*str\s)<<1+2)
    If result
      PokeS(result, *str\s, -1, #PB_Unicode)
    EndIf
  CompilerEndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-comDocDocumentObject CLASS.
;-===========================

;-CLASS CREATION FUNCTIONS.

;/////////////////////////////////////////////////////////////////////////////////
;The following function opens an existing memory-based OLE compound file (based on a memory buffer) and returns, if successful, a
;newly instantiated comDocObject.
;Save the document using the \SaveDocumentAs() method as required.
Procedure.i comDoc_CatchDocument(memoryAddress, size)
  Protected hResult, *this.comDocDocumentObject_ClassTemplate, hGlobal, gPtr
  If memoryAddress And size > 0
    ;Allocate memory for a new class template.
      *this = AllocateMemory(SizeOf(comDocDocumentObject_ClassTemplate))
    If *this
      *this\vTable = ?VTable_documentClass
      ;Create a iLockBytes byte array object housing a copy of the memory buffer.
        hGlobal = GlobalAlloc_(#GMEM_MOVEABLE, size)
        If hGlobal
          ;Copy the memory buffer.
            gPtr = GlobalLock_(hGlobal)
            If gPtr
              CopyMemory(memoryAddress, gPtr, size)
              GlobalUnlock_(hGlobal)
              If CreateILockBytesOnHGlobal_(hGlobal, #True, @*this\lockbytes) = #S_OK
                ;Create our compound storage object atop our iLockBytes array object.
                  If StgOpenStorageOnILockBytes_(*this\lockbytes, 0, #STGM_DIRECT|#STGM_READWRITE|#STGM_SHARE_EXCLUSIVE, 0, 0, @*this\storage) <> #S_OK
                    *this\lockbytes\Release() ;The hGlobal will be freed by this call because of the #True in CreateILockBytesOnHGlobal_().
                    Goto comDoc_CatchDocument_L1
                  EndIf
              Else
                Goto comDoc_CatchDocument_L2
              EndIf
            Else
comDoc_CatchDocument_L2:
              GlobalFree_(hGlobal)
              Goto comDoc_CatchDocument_L1
            EndIf
        Else
comDoc_CatchDocument_L1:
          FreeMemory(*this)
          *this = 0
        EndIf
    EndIf
  EndIf
  ProcedureReturn *this
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following function creates a new disc-based OLE compound file and returns, if successful, a newly instantiated comDocObject.
;If fileName$ is empty then a temporary document is created.
Procedure.i comDoc_CreateDiscDocument(fileName$)
  Protected hResult, uCodeBuffer, *this.comDocDocumentObject_ClassTemplate
  fileName$ = LCase(Trim(fileName$))
  If fileName$
    DeleteFile(fileName$)
    uCodeBuffer = comDoc_INTERNAL_StringToUnicode(@fileName$)
    If uCodeBuffer = 0
      ProcedureReturn 0
    EndIf
  EndIf
  ;Allocate memory for a new class template.
    *this = AllocateMemory(SizeOf(comDocDocumentObject_ClassTemplate))
  If *this
    *this\vTable = ?VTable_documentClass
    ;Try and create a new compound file.
      hResult = StgCreateDocfile_(uCodeBuffer, #STGM_DIRECT|#STGM_CREATE|#STGM_READWRITE|#STGM_SHARE_EXCLUSIVE, 0, @*this\storage)
    If hResult = #S_OK
      *this\fileName$ = fileName$
    Else
      FreeMemory(*this)
      *this = 0
    EndIf 
  EndIf
  If fileName$
    comDoc_INTERNAL_FreeUnicodeString(uCodeBuffer)
  EndIf
  ProcedureReturn *this
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following function creates a new (empty) memory-based OLE compound file and returns, if successful, a newly instantiated
;comDocObject.
;Save the document using the \SaveDocumentAs() method as appropriate.
Procedure.i comDoc_CreateMemoryDocument()
  Protected hResult, *this.comDocDocumentObject_ClassTemplate, hGlobal, gPtr
  ;Allocate memory for a new class template.
    *this = AllocateMemory(SizeOf(comDocDocumentObject_ClassTemplate))
  If *this
    *this\vTable = ?VTable_documentClass
    ;Create a iLockBytes byte array object housing a null global memory object.
      If CreateILockBytesOnHGlobal_(0, #True, @*this\lockbytes) = #S_OK
      ;Create our compound storage object atop our iLockBytes array object.
        If StgCreateDocfileOnILockBytes_(*this\lockbytes, #STGM_DIRECT|#STGM_CREATE|#STGM_READWRITE|#STGM_SHARE_EXCLUSIVE, 0, @*this\storage) <> #S_OK
          *this\lockbytes\Release()
          Goto comDoc_CreateMemoryDocument_L1
        EndIf
      Else
comDoc_CreateMemoryDocument_L1:
        FreeMemory(*this)
        *this = 0
      EndIf
  EndIf
  ProcedureReturn *this
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following function opens an existing disc-based OLE compound file and returns, if successful, a newly instantiated comDocObject.
Procedure.i comDoc_OpenDiscDocument(fileName$)
  Protected hResult, uCodeBuffer, *this.comDocDocumentObject_ClassTemplate
  fileName$ = LCase(Trim(fileName$))
  If fileName$
    uCodeBuffer = comDoc_INTERNAL_StringToUnicode(@fileName$)
    If uCodeBuffer = 0
      ProcedureReturn 0
    EndIf
    ;Allocate memory for a new class template.
      *this = AllocateMemory(SizeOf(comDocDocumentObject_ClassTemplate))
    If *this
      *this\vTable = ?VTable_documentClass
      ;Try and open the compound file.
        hResult = StgOpenStorage_(uCodeBuffer, 0, #STGM_DIRECT_SWMR|#STGM_READWRITE|#STGM_SHARE_EXCLUSIVE, 0, 0, @*this\storage);  
      If hResult = #S_OK
        *this\fileName$ = fileName$
      Else
        FreeMemory(*this)
        *this = 0
      EndIf 
    EndIf
    comDoc_INTERNAL_FreeUnicodeString(uCodeBuffer)
  EndIf
  ProcedureReturn *this
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-CLASS METHODS.

;/////////////////////////////////////////////////////////////////////////////////
Procedure comDoc_documentClass_Close(*this.comDocDocumentObject_ClassTemplate)
  ;First close any active enumeration.
    comDoc_documentClass_FinishEnumeration(*this)  
  ;Next close the compound file.
    *this\storage\Release()
    If *this\lockbytes
      *this\lockbytes\Release()
    EndIf
  ClearStructure(*this, comDocDocumentObject_ClassTemplate)
  FreeMemory(*this)
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method creates a new file (stream object) within the underlying document object and returns a valid comDocFileObject
;pointer if successful.
Procedure.i comDoc_documentClass_CreateFile(*this.comDocDocumentObject_ClassTemplate, name$)
  Protected hResult, *file.comDocFileObject_ClassTemplate, uCodeBuffer
  uCodeBuffer = comDoc_INTERNAL_StringToUnicode(@name$)
  If uCodeBuffer
    ;Allocate memory for a new class template.
      *file = AllocateMemory(SizeOf(comDocFileObject_ClassTemplate))
    If *file
      *file\vTable = ?VTable_fileClass
      ;Try and create a new iStream file object.
        hResult  = *this\storage\CreateStream(uCodeBuffer, #STGM_DIRECT|#STGM_CREATE|#STGM_READWRITE|#STGM_SHARE_EXCLUSIVE, 0, 0, @*file\stream)  
      If hResult = #S_OK 
        *file\name$ = name$
      Else
        FreeMemory(*file)
        *file = 0
      EndIf  
    EndIf
    comDoc_INTERNAL_FreeUnicodeString(uCodeBuffer)
  EndIf
  ProcedureReturn *file
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method deletes an existing file (stream object) from within the underlying document object and returns
;non-zero if successful.
Procedure.i comDoc_documentClass_DeleteFile(*this.comDocDocumentObject_ClassTemplate, name$)
  Protected result, uCodeBuffer
  uCodeBuffer = comDoc_INTERNAL_StringToUnicode(@name$)
  If uCodeBuffer
    If *this\storage\DestroyElement(uCodeBuffer) = #S_OK
      result = #True
    EndIf
    comDoc_INTERNAL_FreeUnicodeString(uCodeBuffer)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method opens an existing file (stream object) within the underlying document object and returns a valid comDocFileObject
;pointer if successful.
Procedure.i comDoc_documentClass_OpenFile(*this.comDocDocumentObject_ClassTemplate, name$)
  Protected hResult, *file.comDocFileObject_ClassTemplate, uCodeBuffer
  uCodeBuffer = comDoc_INTERNAL_StringToUnicode(@name$)
  If uCodeBuffer
    ;Allocate memory for a new class template.
      *file = AllocateMemory(SizeOf(comDocFileObject_ClassTemplate))
    If *file
      *file\vTable = ?VTable_fileClass
      ;Try and create a new iStream file object.
        hResult  = *this\storage\OpenStream(uCodeBuffer, 0, #STGM_DIRECT|#STGM_READWRITE|#STGM_SHARE_EXCLUSIVE, 0, @*file\stream)  
      If hResult = #S_OK 
        *file\name$ = name$
      Else
        FreeMemory(*file)
        *file = 0
      EndIf  
    EndIf
    comDoc_INTERNAL_FreeUnicodeString(uCodeBuffer)
  EndIf
  ProcedureReturn *file
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;Set blnUseNewDocument to non-zero to have the underlying document object reference the newly saved document. In these cases
;any open comDocFileObject objects will be invalidated.
;Returns non-zero if successful.
Procedure.i comDoc_documentClass_SaveDocumentAs(*this.comDocDocumentObject_ClassTemplate, fileName$, blnUseNewDocument=#False)
  Protected result, hResult, storage.IStorage, uCodeBuffer
  fileName$ = LCase(Trim(fileName$))
  If fileName$ <> *this\fileName$
    If fileName$
      DeleteFile(fileName$)
      uCodeBuffer = comDoc_INTERNAL_StringToUnicode(@fileName$)
      If uCodeBuffer = 0
        ProcedureReturn #False
      EndIf
    EndIf
    ;Try and create a new compound file.
      hResult = StgCreateDocfile_(uCodeBuffer, #STGM_DIRECT|#STGM_CREATE|#STGM_READWRITE|#STGM_SHARE_EXCLUSIVE, 0, @storage)
    If hResult = #S_OK
      If *this\storage\CopyTo(0, 0, 0, storage) = #S_OK
        If blnUseNewDocument
          comDoc_documentClass_FinishEnumeration(*this)
          *this\storage\Release()
          If *this\lockbytes
            *this\lockbytes\Release()
            *this\lockbytes = 0
          EndIf
          *this\storage = storage
          *this\fileName$ = fileName$
        Else
          storage\Release()
        EndIf
        result = #True
      Else
        storage\Release()
      EndIf
    EndIf 
    If fileName$
      comDoc_INTERNAL_FreeUnicodeString(uCodeBuffer)
    EndIf
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method begins an enumeration of the documents contents.
;Returns non-zero if successful.
Procedure.i comDoc_documentClass_BeginEnumeration(*this.comDocDocumentObject_ClassTemplate)
  Protected result
  ;If we already have an enumeration in process then we simply reset it.
    If *this\enumerate
      *this\enumerate\Reset()
      result = #True
    Else
      ;First obtain a iMalloc interface pointer in order to free some memory later on.
        If CoGetMalloc_(1, @*this\malloc) = #S_OK
          ;Now attempt to create the enumeration.
            If *this\storage\EnumElements(0, 0, 0, @*this\enumerate) = #S_OK
              result = #True
            Else
              *this\malloc\Release()
              *this\malloc = 0
            EndIf
        EndIf
    EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.s comDoc_documentClass_EnumerationEntryName(*this.comDocDocumentObject_ClassTemplate)
  Protected result$
  If *this\enumerate And *this\statstg\pwcsName
    result$ = PeekS(*this\statstg\pwcsName, -1, #PB_Unicode)
  EndIf
  ProcedureReturn result$
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i comDoc_documentClass_EnumerationEntrySize(*this.comDocDocumentObject_ClassTemplate)
  Protected result
  If *this\enumerate
    result = *this\statstg\cbSize
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i comDoc_documentClass_EnumerationEntryType(*this.comDocDocumentObject_ClassTemplate)
  Protected result
  If *this\enumerate
    result = *this\statstg\type
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method finishes an enumeration of the documents contents.
;Returns non-zero if successful.
Procedure comDoc_documentClass_FinishEnumeration(*this.comDocDocumentObject_ClassTemplate)
  If *this\enumerate
    ;Free any STATSTG structure memory.
      If *this\statstg\pwcsName
        *this\malloc\Free(*this\statstg\pwcsName)
      EndIf
    *this\malloc\Release()
    *this\malloc = 0
    *this\enumerate\Release()
    *this\enumerate = 0
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method advances a document enumeration to the next entry.
;Returns non-zero if successful.
Procedure.i comDoc_documentClass_NextEnumeration(*this.comDocDocumentObject_ClassTemplate)
  Protected result, numStatstg
  If *this\enumerate
    ;Free any existing entry.
      If *this\statstg\pwcsName
        *this\malloc\Free(*this\statstg\pwcsName)
      EndIf
    If *this\enumerate\Next(1, @*this\statstg, @numStatstg) = #S_OK And numStatstg = 1
      result = #True
    EndIf
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;->>
;-comDocFileObject CLASS.
;-=======================

;-CLASS METHODS.

;/////////////////////////////////////////////////////////////////////////////////
Procedure comDoc_fileClass_Close(*this.comDocFileObject_ClassTemplate)
  ;First close the file.
    *this\stream\Release()
  ClearStructure(*this, comDocFileObject_ClassTemplate)
  FreeMemory(*this)
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i comDoc_fileClass_Eof(*this.comDocFileObject_ClassTemplate)
  Protected result, loc, lof
  *this\stream\Seek(0, #STREAM_SEEK_CUR, @loc)
  *this\stream\Seek(0, #STREAM_SEEK_END, @lof)
  *this\stream\Seek(loc, #STREAM_SEEK_SET, 0)
  If loc >= lof
    result = #True
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i comDoc_fileClass_FileSeek(*this.comDocFileObject_ClassTemplate, newPosition)
  *this\stream\Seek(newPosition, #STREAM_SEEK_SET, 0)
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i comDoc_fileClass_Loc(*this.comDocFileObject_ClassTemplate)
  Protected result
  *this\stream\Seek(0, #STREAM_SEEK_CUR, @result)
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i comDoc_fileClass_Lof(*this.comDocFileObject_ClassTemplate)
  Protected result, current
  *this\stream\Seek(0, #STREAM_SEEK_CUR, @current)
  *this\stream\Seek(0, #STREAM_SEEK_END, @result)
  *this\stream\Seek(current, #STREAM_SEEK_SET, 0)
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.a comDoc_fileClass_ReadAsciiCharacter(*this.comDocFileObject_ClassTemplate)
  Protected buffer.a
  *this\stream\Read(@buffer, SizeOf(ASCII), 0)
  ProcedureReturn buffer
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.b comDoc_fileClass_ReadByte(*this.comDocFileObject_ClassTemplate)
  Protected buffer.b
  *this\stream\Read(@buffer, SizeOf(BYTE), 0)
  ProcedureReturn buffer
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.c comDoc_fileClass_ReadCharacter(*this.comDocFileObject_ClassTemplate)
  Protected buffer.c
  *this\stream\Read(@buffer, SizeOf(CHARACTER), 0)
  ProcedureReturn buffer
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;Returns the number of bytes actually read.
Procedure.i comDoc_fileClass_ReadData(*this.comDocFileObject_ClassTemplate, *memoryBuffer, lengthToRead)
  Protected numBytesRead
  If *memoryBuffer And lengthToRead
    *this\stream\Read(*memoryBuffer, lengthToRead, @numBytesRead)
  EndIf
  ProcedureReturn numBytesRead
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.d comDoc_fileClass_ReadDouble(*this.comDocFileObject_ClassTemplate)
  Protected buffer.d
  *this\stream\Read(@buffer, SizeOf(DOUBLE), 0)
  ProcedureReturn buffer
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.f comDoc_fileClass_ReadFloat(*this.comDocFileObject_ClassTemplate)
  Protected buffer.f
  *this\stream\Read(@buffer, SizeOf(FLOAT), 0)
  ProcedureReturn buffer
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i comDoc_fileClass_ReadInteger(*this.comDocFileObject_ClassTemplate)
  Protected buffer.i
  *this\stream\Read(@buffer, SizeOf(INTEGER), 0)
  ProcedureReturn buffer
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.l comDoc_fileClass_ReadLong(*this.comDocFileObject_ClassTemplate)
  Protected buffer.l
  *this\stream\Read(@buffer, SizeOf(LONG), 0)
  ProcedureReturn buffer
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.q comDoc_fileClass_ReadQuad(*this.comDocFileObject_ClassTemplate)
  Protected buffer.q
  *this\stream\Read(@buffer, SizeOf(QUAD), 0)
  ProcedureReturn buffer
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.s comDoc_fileClass_ReadString(*this.comDocFileObject_ClassTemplate, format=0)
  Protected text$, buffer, byteLen, loc, lof, t1, byte, char, numBytes, charSize
  If format <> #PB_Ascii And format <> #PB_UTF8 And format <> #PB_Unicode
    CompilerIf #PB_Compiler_Unicode
      format = #PB_Unicode
    CompilerElse
      format = #PB_Ascii
    CompilerEndIf
  EndIf
  If format = #PB_Unicode
    charSize = 2
  Else
    charSize = 1
  EndIf
  ;Determine the number of bytes to read.
    *this\stream\Seek(0, #STREAM_SEEK_CUR, @loc)
    *this\stream\Seek(0, #STREAM_SEEK_END, @lof)
    *this\stream\Seek(loc, #STREAM_SEEK_SET, 0)
    t1 = loc
    While t1 < lof
      char = 0
      *this\stream\Read(@char, charSize, @numBytes)     
      byteLen+numBytes
      If char = 0
        Break
      EndIf
      t1+numBytes
    Wend
    *this\stream\Seek(loc, #STREAM_SEEK_SET, 0)
  If byteLen
    buffer = AllocateMemory(byteLen+2)  ;+2 in case the string is not null-terminated and we hit the end of the file.
    If buffer
      If *this\stream\Read(buffer, byteLen, @numBytes) = #S_OK And numBytes = byteLen
        text$ = PeekS(buffer, -1, format)
      EndIf
      FreeMemory(buffer)
    EndIf
  EndIf
  ProcedureReturn text$
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.u comDoc_fileClass_ReadUnicodeCharacter(*this.comDocFileObject_ClassTemplate)
  Protected buffer.u
  *this\stream\Read(@buffer, SizeOf(UNICODE), 0)
  ProcedureReturn buffer
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.w comDoc_fileClass_ReadWord(*this.comDocFileObject_ClassTemplate)
  Protected buffer.w
  *this\stream\Read(@buffer, SizeOf(WORD), 0)
  ProcedureReturn buffer
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure comDoc_fileClass_WriteAsciiCharacter(*this.comDocFileObject_ClassTemplate, value.a)
  Protected numBytes
  If *this\stream\Write(@value, SizeOf(ASCII), @numBytes) <> #S_OK Or numbytes <> SizeOf(ASCII)
    *this\errorFlag = #True
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure comDoc_fileClass_WriteByte(*this.comDocFileObject_ClassTemplate, value.b)
  Protected numBytes
  If *this\stream\Write(@value, SizeOf(BYTE), @numBytes) <> #S_OK Or numbytes <> SizeOf(BYTE)
    *this\errorFlag = #True
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure comDoc_fileClass_WriteCharacter(*this.comDocFileObject_ClassTemplate, value.c)
  Protected numBytes
  If *this\stream\Write(@value, SizeOf(CHARACTER), @numBytes) <> #S_OK Or numbytes <> SizeOf(CHARACTER)
    *this\errorFlag = #True
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure comDoc_fileClass_WriteData(*this.comDocFileObject_ClassTemplate, *memoryBuffer, length)
  Protected numBytes
  If *memoryBuffer And length
    If *this\stream\Write(*memoryBuffer, length, @numBytes) <> #S_OK Or numbytes <> length
      *this\errorFlag = #True
    EndIf
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure comDoc_fileClass_WriteDouble(*this.comDocFileObject_ClassTemplate, value.d)
  Protected numBytes
  If *this\stream\Write(@value, SizeOf(DOUBLE), @numBytes) <> #S_OK Or numbytes <> SizeOf(DOUBLE)
    *this\errorFlag = #True
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure comDoc_fileClass_WriteFloat(*this.comDocFileObject_ClassTemplate, value.f)
  Protected numBytes
  If *this\stream\Write(@value, SizeOf(FLOAT), @numBytes) <> #S_OK Or numbytes <> SizeOf(FLOAT)
    *this\errorFlag = #True
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure comDoc_fileClass_WriteInteger(*this.comDocFileObject_ClassTemplate, value.i)
  Protected numBytes
  If *this\stream\Write(@value, SizeOf(INTEGER), @numBytes) <> #S_OK Or numbytes <> SizeOf(INTEGER)
    *this\errorFlag = #True
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure comDoc_fileClass_WriteLong(*this.comDocFileObject_ClassTemplate, value.l)
  Protected numBytes
  If *this\stream\Write(@value, SizeOf(LONG), @numBytes) <> #S_OK Or numbytes <> SizeOf(LONG)
    *this\errorFlag = #True
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure comDoc_fileClass_WriteQuad(*this.comDocFileObject_ClassTemplate, value.q)
  Protected numBytes
  If *this\stream\Write(@value, SizeOf(QUAD), @numBytes) <> #S_OK Or numbytes <> SizeOf(QUAD)
    *this\errorFlag = #True
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure comDoc_fileClass_WriteString(*this.comDocFileObject_ClassTemplate, text$, format=0)
  Protected numBytes, byteLength, buffer
  If format <> #PB_Ascii And format <> #PB_UTF8 And format <> #PB_Unicode
    CompilerIf #PB_Compiler_Unicode
      format = #PB_Unicode
    CompilerElse
      format = #PB_Ascii
    CompilerEndIf
  EndIf
  byteLength = StringByteLength(text$+"a", format) ;The 'a' takes the place of the terminating null(s).
  buffer = AllocateMemory(byteLength)
  If buffer
    PokeS(buffer, text$, -1, format)  
    If *this\stream\Write(buffer, byteLength, @numBytes) <> #S_OK Or numbytes <> byteLength
      *this\errorFlag = #True
    EndIf
    FreeMemory(buffer)
  Else
    *this\errorFlag = #True
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure comDoc_fileClass_WriteUnicodeCharacter(*this.comDocFileObject_ClassTemplate, value.u)
  Protected numBytes
  If *this\stream\Write(@value, SizeOf(UNICODE), @numBytes) <> #S_OK Or numbytes <> SizeOf(UNICODE)
    *this\errorFlag = #True
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure comDoc_fileClass_WriteWord(*this.comDocFileObject_ClassTemplate, value.w)
  Protected numBytes
  If *this\stream\Write(@value, SizeOf(WORD), @numBytes) <> #S_OK Or numbytes <> SizeOf(WORD)
    *this\errorFlag = #True
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure comDoc_fileClass_ClearErrorFlag(*this.comDocFileObject_ClassTemplate)
  *this\errorFlag = #False
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i comDoc_fileClass_GetErrorFlag(*this.comDocFileObject_ClassTemplate)
  ProcedureReturn *this\errorFlag
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


DataSection

  VTable_documentClass:
    Data.i @comDoc_documentClass_Close()
    Data.i @comDoc_documentClass_CreateFile()
    Data.i @comDoc_documentClass_DeleteFile()
    Data.i @comDoc_documentClass_OpenFile()
    Data.i @comDoc_documentClass_SaveDocumentAs()
    Data.i @comDoc_documentClass_BeginEnumeration()
    Data.i @comDoc_documentClass_EnumerationEntryName()
    Data.i @comDoc_documentClass_EnumerationEntrySize()
    Data.i @comDoc_documentClass_EnumerationEntryType()
    Data.i @comDoc_documentClass_FinishEnumeration()
    Data.i @comDoc_documentClass_NextEnumeration()

  VTable_fileClass:
    Data.i @comDoc_fileClass_Close()
    Data.i @comDoc_fileClass_Eof()
    Data.i @comDoc_fileClass_FileSeek()
    Data.i @comDoc_fileClass_Loc()
    Data.i @comDoc_fileClass_Lof()
    Data.i @comDoc_fileClass_ReadAsciiCharacter()
    Data.i @comDoc_fileClass_ReadByte()
    Data.i @comDoc_fileClass_ReadCharacter()
    Data.i @comDoc_fileClass_ReadData()
    Data.i @comDoc_fileClass_ReadDouble()
    Data.i @comDoc_fileClass_ReadFloat()
    Data.i @comDoc_fileClass_ReadInteger()
    Data.i @comDoc_fileClass_ReadLong()
    Data.i @comDoc_fileClass_ReadQuad()
    Data.i @comDoc_fileClass_ReadString()
    Data.i @comDoc_fileClass_ReadUnicodeCharacter()
    Data.i @comDoc_fileClass_ReadWord()
    Data.i @comDoc_fileClass_WriteAsciiCharacter()
    Data.i @comDoc_fileClass_WriteByte()
    Data.i @comDoc_fileClass_WriteCharacter()
    Data.i @comDoc_fileClass_WriteData()
    Data.i @comDoc_fileClass_WriteDouble()
    Data.i @comDoc_fileClass_WriteFloat()
    Data.i @comDoc_fileClass_WriteInteger()
    Data.i @comDoc_fileClass_WriteLong()
    Data.i @comDoc_fileClass_WriteQuad()
    Data.i @comDoc_fileClass_WriteString()
    Data.i @comDoc_fileClass_WriteUnicodeCharacter()
    Data.i @comDoc_fileClass_WriteWord()
    Data.i @comDoc_fileClass_ClearErrorFlag()
    Data.i @comDoc_fileClass_GetErrorFlag()
EndDataSection

CompilerEndIf



; IDE Options = PureBasic 4.51 (Windows - x86)
; ExecutableFormat = Shared Dll
; CursorPosition = 489
; FirstLine = 432
; Folding = ---------
; Markers = 352,400
; EnableUnicode
; EnableXP
; Executable = nxReportU.dll